/*
  Copyright (c) 2013, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include "mmss.hpp"

#include <cassert>
#include <cstdio>


namespace MMSS {

static void init_nodes(std::list<std::shared_ptr<node_t>> &nodes) {
	for (auto node : nodes)
		node->ctx = node->proto->init(node.get());
}


static inline int timeout_min(int a, int b) {
	if (a < 0)
		return b;
	else if (b < 0)
		return a;
	else
		return min(a, b);
}


static int get_queue_timeout(const context_t *mmss) {
	return timeout_min(mmss->packet_queue.timeout(), mmss->scheduled_queue.timeout());
}

uint64_t now(const context_t *mmss) {
	return mmss->now;
}

void main(int argc, char *argv[]) {
	if (argc != 2) {
		std::fprintf(stderr, "usage: %s protocol_module\n", argv[0]);
		std::exit(1);
	}

	const protocol_t *proto = load_protocol(argv[1]);
	if (!proto)
		std::exit(1);

	context_t mmss;
	config_t conf = {};

	//read_config(&mmss, &conf, "babel_test.mmss");

	std::shared_ptr<network_t> net0 = std::make_shared<network_t>();
	std::shared_ptr<network_t> net1 = std::make_shared<network_t>();

	net0->mtu = 1500;
	net1->mtu = 1500;

	std::shared_ptr<node_t> node1 = node_t::create(&mmss, "node1", 1, proto);
	std::shared_ptr<node_t> node2 = node_t::create(&mmss, "node2", 2, proto);
	std::shared_ptr<node_t> node3 = node_t::create(&mmss, "node3", 3, proto);

	std::list<std::shared_ptr<node_t>> nodes;

	nodes.push_back(node1);
	nodes.push_back(node2);
	nodes.push_back(node3);

	init_nodes(nodes);

	gmrf_addr_t addr1 = {{1}}, addr2 = {{2}}, addr3 = {{3}}, addr4 = {{4}};
	add_iface(node1, net0, "mmss0", &addr1);
	add_iface(node2, net0, "mmss0", &addr2);
	add_iface(node2, net1, "mmss1", &addr3);
	add_iface(node3, net1, "mmss1", &addr4);

	while (true) {
		if (mmss.now > 1000000000)
			break;

		int timeout = get_queue_timeout(&mmss);

		if (timeout < 0) {
			fprintf(stderr, "nothing queued, deadlock occured.\n");
			break;
		}

		if (timeout > 0) {
			assert(!mmss.packet_queue.get());
			assert(!mmss.scheduled_queue.get());

			mmss.now += timeout;
			timeout = get_queue_timeout(&mmss);
		}

		assert(timeout == 0);

		while (timeout == 0) {
			std::shared_ptr<packet_t> packet = mmss.packet_queue.get();
			std::shared_ptr<scheduled_t> scheduled = mmss.scheduled_queue.get();

			assert(packet || scheduled);

			if (packet)
				dispatch(packet);

			if (scheduled)
				run_scheduled(scheduled);

			timeout = get_queue_timeout(&mmss);
		}
	}
}

}

int main(int argc, char *argv[]) {
	MMSS::main(argc, argv);

	return 0;
}
