/*
  Copyright (c) 2013, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#ifndef _GMRF_MMSS_MMSS_HPP_
#define _GMRF_MMSS_MMSS_HPP_

#include "queue.hpp"

#include <list>
#include <memory>


struct gmrf {
private:
	gmrf(MMSS::context_t *mmss0, const std::string &name0, unsigned rand_seed0, const MMSS::protocol_t *proto0) :
		mmss(mmss0), name(name0), rand_seed(rand_seed0), proto(proto0) {}
	gmrf(gmrf const&) = delete;
	gmrf& operator=(gmrf const&) = delete;

public:
	static std::shared_ptr<MMSS::node_t> create(MMSS::context_t *mmss, const std::string &name, unsigned rand_seed, const MMSS::protocol_t *proto) {
		MMSS::node_t *node = new MMSS::node_t(mmss, name, rand_seed, proto);

		std::shared_ptr<MMSS::node_t> ptr(node);
		node->node = ptr;

		return ptr;
	}

	std::weak_ptr<MMSS::node_t> node;

	MMSS::context_t *mmss;

	std::string name;

	gmrf_context_t *ctx;
	std::list<std::shared_ptr<MMSS::iface_t>> interfaces;

	unsigned rand_seed;

	const MMSS::protocol_t *proto;
};

struct gmrf_iface {
private:
	gmrf_iface() {}
	gmrf_iface(gmrf_iface const&) = delete;
	gmrf_iface& operator=(gmrf_iface const&) = delete;

public:
	static std::shared_ptr<MMSS::iface_t> create() {
		MMSS::iface_t *iface = new MMSS::iface_t;

		std::shared_ptr<MMSS::iface_t> ptr(iface);
		iface->iface = ptr;

		return ptr;
	}

	std::weak_ptr<MMSS::iface_t> iface;

	std::string name;
	gmrf_addr_t address;

	std::weak_ptr<MMSS::node_t> node;
	std::weak_ptr<MMSS::network_t> net;
};


namespace MMSS {

class context_t {
public:
	uint64_t now;
	timeout_queue_t<packet_t> packet_queue;
	timeout_queue_t<scheduled_t> scheduled_queue;

	context_t() : now(0), packet_queue(this), scheduled_queue(this) {}
};

class config_t {
public:
	std::list<std::shared_ptr<network_t>> network;
	std::list<std::shared_ptr<node_t>> nodes;
};

class network_t {
public:
	std::string name;

	std::list<std::shared_ptr<iface_t>> interfaces;
	size_t mtu;
};

class packet_t {
public:
	uint64_t sent;

	std::weak_ptr<iface_t> source;
	std::weak_ptr<iface_t> dest;

	size_t len;
	std::unique_ptr<uint8_t> data;
};

class scheduled_t {
public:
	std::shared_ptr<node_t> node;
	gmrf_scheduled_func f;
	void *arg;
};


const protocol_t* load_protocol(const char *module);

bool read_config(context_t *mmss, config_t *conf, const char *filename);

void add_iface(const std::shared_ptr<node_t> &node, const std::shared_ptr<network_t> &net, const std::string &name, const gmrf_addr_t *address);

uint64_t now(const context_t *mmss);

void dispatch(const std::shared_ptr<packet_t> &packet);
void run_scheduled(const std::shared_ptr<scheduled_t> &scheduled);
void enqueue(context_t *mmss, const std::shared_ptr<iface_t> &source, const std::shared_ptr<iface_t> &dest, const void *data, size_t len);

void logf(context_t *mmss, int priority, const char *format, ...);


static inline int max(int a, int b) {
	return (a > b) ? a : b;
}

static inline int min(int a, int b) {
	return (a < b) ? a : b;
}


static inline size_t alignto(size_t l, size_t a) {
	return ((l+a-1)/a)*a;
}


namespace Config {

void add_network(context_t *mmss, config_t *conf, const char *name);

}

}


#endif /* _GMRF_MMSS_MMSS_HPP_ */
